package com.microduino.flutter_js_bridge;

import android.content.Context;
import android.util.Log;
import android.view.View;
import android.webkit.WebStorage;

import com.github.lzyzsd.jsbridge.BridgeHandler;
import com.github.lzyzsd.jsbridge.BridgeWebView;
import com.github.lzyzsd.jsbridge.CallBackFunction;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import io.flutter.plugin.common.BinaryMessenger;
import io.flutter.plugin.common.MethodCall;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.platform.PlatformView;

/**
 * Created by Gloria Guo on ${Date}.
 */
public class FlutterBridgeWebView implements PlatformView, MethodChannel.MethodCallHandler {
    private final String tag = "FlutterBridgeWebView";
    private final BridgeWebView webView;
    private final MethodChannel methodChannel;
    private final FlutterBridgeWebViewClient webViewClient;

    @SuppressWarnings("unchecked")
    FlutterBridgeWebView(Context context, BinaryMessenger messenger, int id, Map<String, Object> params) {
        webView = new BridgeWebView(context);
        // Allow local storage.
        webView.getSettings().setDomStorageEnabled(true);
        webView.getSettings().setBuiltInZoomControls(false);
        webView.getSettings().setUseWideViewPort(false);

        methodChannel = new MethodChannel(messenger, "com.microduino.flutter/bridgewebview_" + id);
        methodChannel.setMethodCallHandler(this);

        webViewClient = new FlutterBridgeWebViewClient(webView, methodChannel);
        applySettings((Map<String, Object>) params.get("settings"));

        webView.setWebViewClient(webViewClient);

        if (params.containsKey("bridgeChannelNames")) {
            registerBridgeChannelNames((List<String>) params.get("bridgeChannelNames"));
        }

        if (params.containsKey("initialUrl")) {
            String url = (String) params.get("initialUrl");
            webView.loadUrl(url);
        }

    }

    @Override
    public View getView() {
        return webView;
    }

    @Override
    public void onMethodCall(MethodCall methodCall, MethodChannel.Result result) {
        switch (methodCall.method) {
            case "loadUrl":
                loadUrl(methodCall, result);
                break;
            case "updateSettings":
                updateSettings(methodCall, result);
                break;
            case "canGoBack":
                canGoBack(methodCall, result);
                break;
            case "canGoForward":
                canGoForward(methodCall, result);
                break;
            case "goBack":
                goBack(methodCall, result);
                break;
            case "goForward":
                goForward(methodCall, result);
                break;
            case "reload":
                reload(methodCall, result);
                break;
            case "currentUrl":
                currentUrl(methodCall, result);
                break;
            case "callHandler":
                callHandler(methodCall, result);
                break;
            case "registerHandler":
                registerHandler(methodCall, result);
                break;
            case "clearCache":
                clearCache(result);
                break;
            default:
                result.notImplemented();
        }
    }

    private void loadUrl(MethodCall methodCall, MethodChannel.Result result) {
        String url = (String) methodCall.arguments;
        webView.loadUrl(url);
        result.success(null);
    }

    private void registerBridgeChannelNames(List<String> channelNames) {
        for (String channelName : channelNames) {
            webView.registerHandler(channelName, new BridgeHandler() {
                @Override
                public void handler(String data, CallBackFunction function) {
                    Log.e(tag, "------> GET registerHandler callback event");
                    methodChannel.invokeMethod("onRegisterHandlerCallback", new HashMap<>());
                }
            });
        }
    }

    private void canGoBack(MethodCall methodCall, MethodChannel.Result result) {
        result.success(webView.canGoBack());
    }

    private void canGoForward(MethodCall methodCall, MethodChannel.Result result) {
        result.success(webView.canGoForward());
    }

    private void goBack(MethodCall methodCall, MethodChannel.Result result) {
        if (webView.canGoBack()) {
            webView.goBack();
        }
        result.success(null);
    }

    private void goForward(MethodCall methodCall, MethodChannel.Result result) {
        if (webView.canGoForward()) {
            webView.goForward();
        }
        result.success(null);
    }

    private void reload(MethodCall methodCall, MethodChannel.Result result) {
        webView.reload();
        result.success(null);
    }

    private void currentUrl(MethodCall methodCall, MethodChannel.Result result) {
        result.success(webView.getUrl());
    }

    private void callHandler(MethodCall methodCall, MethodChannel.Result result) {
        final Map<String, Object> method = (Map<String, Object>) methodCall.arguments;
        final String name = (String) method.get("name");
        final String data = (String) method.get("data");
        Log.e(tag, "----> callHandler [" + method.get("name") + "(" + data + ")]");
        webView.callHandler(name, data, new CallBackFunction() {
            @Override
            public void onCallBack(String data) {
                Map<String, Object> args = new HashMap<>();
                args.put("data", data);
                args.put("name", name);
                methodChannel.invokeMethod("onCallBack", args);
            }
        });
        result.success(null);
    }

    private void registerHandler(MethodCall methodCall, MethodChannel.Result result) {
        final String methodName = (String) methodCall.arguments;
        Log.e(tag, "----> registerHandler [" + methodName + "]");
        webView.registerHandler(methodName, new BridgeHandler() {
            @Override
            public void handler(String data, CallBackFunction callBackFunction) {
                callBackFunction.onCallBack(data);
            }
        });
        result.success(null);
    }

    @SuppressWarnings("unchecked")
    private void updateSettings(MethodCall methodCall, MethodChannel.Result result) {
        applySettings((Map<String, Object>) methodCall.arguments);
        result.success(null);
    }

    private void clearCache(MethodChannel.Result result) {
        webView.clearCache(true);
        WebStorage.getInstance().deleteAllData();
        result.success(null);
    }

    private void applySettings(Map<String, Object> settings) {
        for (String key : settings.keySet()) {
            switch (key) {
                case "jsMode":
                    updateJsMode((Integer) settings.get(key));
                    break;
                default:
                    throw new IllegalArgumentException("Unknown WebView setting: " + key);
            }
        }
    }

    private void updateJsMode(int mode) {
        switch (mode) {
            case 0: // disabled
                webView.getSettings().setJavaScriptEnabled(false);
                break;
            case 1: // unrestricted
                webView.getSettings().setJavaScriptEnabled(true);
                break;
            default:
                throw new IllegalArgumentException("Trying to set unknown JavaScript mode: " + mode);
        }
    }

    @Override
    public void dispose() {
        methodChannel.setMethodCallHandler(null);
    }
}
