
import 'dart:async';

import 'package:flutter/foundation.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/services.dart';
import 'package:flutter/widgets.dart';

typedef void WebViewCreatedCallback(BridgeWebViewController controller);

enum JavascriptMode {
    disabled,
    unrestricted,
}

typedef void PageFinishedCallback(String url);
typedef void BridgeHandlerCallback();

final RegExp _validChannelNames = RegExp('^[a-zA-Z_][a-zA-Z0-9]*\$');

class BridgeWebView extends StatefulWidget {
    const BridgeWebView({
        Key key,
        this.onWebViewCreated,
        this.initialUrl,
        this.javascriptMode = JavascriptMode.disabled,
        this.gestureRecognizers,
        this.onPageFinished,
        this.bridgeChannels,
    })
        : assert(javascriptMode != null),
            super(key: key);

    final WebViewCreatedCallback onWebViewCreated;
    final Set<Factory<OneSequenceGestureRecognizer>> gestureRecognizers;
    final String initialUrl;
    final JavascriptMode javascriptMode;
    final PageFinishedCallback onPageFinished;
    final Set<BridgeChannel> bridgeChannels;

    @override
    State<StatefulWidget> createState() => _BridgeWebViewState();
}

class _BridgeWebViewState extends State<BridgeWebView> {
    final Completer<BridgeWebViewController> _controller =
    Completer<BridgeWebViewController>();

    _WebSettings _settings;

    @override
    Widget build(BuildContext context) {
        if (defaultTargetPlatform == TargetPlatform.android) {
            return GestureDetector(
                onLongPress: () {},
                excludeFromSemantics: true,
                child: AndroidView(
                    viewType: 'com.microduino.flutter/bridgewebview',
                    onPlatformViewCreated: _onPlatformViewCreated,
                    gestureRecognizers: widget.gestureRecognizers,
                    // WebView content is not affected by the Android view's layout direction,
                    // we explicitly set it here so that the widget doesn't require an ambient
                    // directionality.
                    layoutDirection: TextDirection.rtl,
                    creationParams: _CreationParams.fromWidget(widget).toMap(),
                    creationParamsCodec: const StandardMessageCodec(),
                ),
            );
        } else if (defaultTargetPlatform == TargetPlatform.iOS) {
            return UiKitView(
                viewType: 'com.microduino.flutter/bridgewebview',
                onPlatformViewCreated: _onPlatformViewCreated,
                gestureRecognizers: widget.gestureRecognizers,
                creationParams: _CreationParams.fromWidget(widget).toMap(),
                creationParamsCodec: const StandardMessageCodec(),
            );
        }
        return Text(
            '$defaultTargetPlatform is not yet supported by the webview_flutter plugin');
    }

    @override
    void initState() {
        super.initState();
    }

    @override
    void didUpdateWidget(BridgeWebView oldWidget) {
        super.didUpdateWidget(oldWidget);
        _assertBridgeChannelNamesAreUnique();
        _updateConfiguration(_WebSettings.fromWidget(widget));
    }

    Future<void> _updateConfiguration(_WebSettings settings) async {
        _settings = settings;
        final BridgeWebViewController controller = await _controller.future;
        controller._updateSettings(settings);
        controller._updateBridgeChannels(widget.bridgeChannels);
    }

    void _onPlatformViewCreated(int id) {
        final BridgeWebViewController controller = BridgeWebViewController._(
            id,
            _WebSettings.fromWidget(widget),
            widget.onPageFinished,
            widget.bridgeChannels
        );
        _controller.complete(controller);
        if (widget.onWebViewCreated != null) {
            widget.onWebViewCreated(controller);
        }
    }

    void _assertBridgeChannelNamesAreUnique() {
        if (widget.bridgeChannels == null ||
            widget.bridgeChannels.isEmpty) {
            return;
        }
        assert(_extractChannelNames(widget.bridgeChannels).length ==
            widget.bridgeChannels.length);
    }

}

Set<String> _extractChannelNames(Set<BridgeChannel> channels) {
    final Set<String> channelNames = channels == null
        ? Set<String>()
        : channels.map((BridgeChannel channel) => channel.name).toSet();
    return channelNames;
}

class _CreationParams {
    _CreationParams(
        {this.initialUrl, this.settings, this.bridgeChannelNames});

    static _CreationParams fromWidget(BridgeWebView widget) {
        return _CreationParams(
            initialUrl: widget.initialUrl,
            settings: _WebSettings.fromWidget(widget),
            bridgeChannelNames: _extractChannelNames(widget.bridgeChannels).toList(),
        );
    }

    final String initialUrl;
    final _WebSettings settings;
    final List<String> bridgeChannelNames;

    Map<String, dynamic> toMap() {
        return <String, dynamic>{
            'initialUrl': initialUrl,
            'settings': settings.toMap(),
            'bridgeChannelNames': bridgeChannelNames,
        };
    }
}

class _WebSettings {
    _WebSettings({
        this.javascriptMode
    });

    static _WebSettings fromWidget(BridgeWebView widget) {
        return _WebSettings(
            javascriptMode: widget.javascriptMode,
        );
    }

    final JavascriptMode javascriptMode;

    Map<String, dynamic> toMap() {
        return <String, dynamic>{
            'jsMode': javascriptMode.index,
        };
    }

    Map<String, dynamic> updatesMap(_WebSettings newSettings) {
        final Map<String, dynamic> updates = <String, dynamic>{};
        if (javascriptMode != newSettings.javascriptMode) {
            updates['jsMode'] = newSettings.javascriptMode.index;
        }
        return updates;
    }
}

class BridgeChannel {
    BridgeChannel({
        @required this.name,
        @required this.onBridgeHandler,
    })
        : assert(name != null),
            assert(onBridgeHandler != null),
            assert(_validChannelNames.hasMatch(name));

    final String name;
    final BridgeHandlerCallback onBridgeHandler;
}

class BridgeWebViewController {
    BridgeWebViewController._(int id,
        this._settings,
        this._onPageFinished,
        Set<BridgeChannel> bridgeChannels)
        : _channel = MethodChannel('com.microduino.flutter/bridgewebview_$id') {
        _updateBridgeChannelsFromSet(bridgeChannels);
        _channel.setMethodCallHandler(_onMethodCall);
    }

    final MethodChannel _channel;

    _WebSettings _settings;

    final PageFinishedCallback _onPageFinished;

    Map<String, BridgeChannel> _bridgeChannels = <String, BridgeChannel>{};

    void _updateBridgeChannelsFromSet(Set<BridgeChannel> channels) {
        _bridgeChannels.clear();
        if (channels == null) {
            return;
        }
        for (BridgeChannel channel in channels) {
            _bridgeChannels[channel.name] = channel;
        }
    }

    Future<bool> _onMethodCall(MethodCall call) async {
        switch (call.method) {
            case 'onPageFinished':
                if (_onPageFinished != null) {
                    _onPageFinished(call.arguments['url']);
                }
                return null;
            case 'onCallBack':
                print('----> onCallBack ${call.arguments}');
                return null;
            case 'onBridgeHandler':
                final String channel = call.arguments['channel'];
                _bridgeChannels[channel].onBridgeHandler();
                return true;
        }
        throw MissingPluginException(
            '${call.method} was invoked but has no handler');
    }

    Future<void> loadUrl(String url) async {
        assert(url != null);
        _validateUrlString(url);
        return _channel.invokeMethod('loadUrl', url);
    }

    Future<void> callHandler(String method, String data) async {
        assert(method != null);
        return _channel.invokeMethod(
            'callHandler', {"name": method, "data": data});
    }

    Future<void> registerHandler(String method) async {
        assert(method != null);
        return _channel.invokeMethod('registerHandler', method);
    }

    Future<String> currentUrl() async {
        final String url = await _channel.invokeMethod('currentUrl');
        return url;
    }

    Future<bool> canGoBack() async {
        final bool canGoBack = await _channel.invokeMethod("canGoBack");
        return canGoBack;
    }

    Future<bool> canGoForward() async {
        final bool canGoForward = await _channel.invokeMethod("canGoForward");
        return canGoForward;
    }

    Future<void> goBack() async {
        return _channel.invokeMethod("goBack");
    }

    Future<void> goForward() async {
        return _channel.invokeMethod("goForward");
    }

    Future<void> reload() async {
        return _channel.invokeMethod("reload");
    }

    Future<void> clearCache() async {
        await _channel.invokeMethod("clearCache");
        return reload();
    }

    Future<void> _updateSettings(_WebSettings setting) async {
        final Map<String, dynamic> updateMap = _settings.updatesMap(setting);
        if (updateMap == null || updateMap.isEmpty) {
            return null;
        }
        _settings = setting;
        return _channel.invokeMethod('updateSettings', updateMap);
    }

    Future<void> _updateBridgeChannels(Set<BridgeChannel> newChannels) async {
        final Set<String> currentChannels = _bridgeChannels.keys.toSet();
        final Set<String> newChannelNames = _extractChannelNames(newChannels);
        final Set<String> channelsToAdd = newChannelNames.difference(currentChannels);
        final Set<String> channelsToRemove = currentChannels.difference(newChannelNames);
        if (channelsToRemove.isNotEmpty) {
            _channel.invokeMethod(
                'removeBridgeChannels', channelsToRemove.toList());
        }
        if (channelsToAdd.isNotEmpty) {
            _channel.invokeMethod(
                'addBridgeChannels', channelsToAdd.toList());
        }
        _updateBridgeChannelsFromSet(newChannels);
    }

}

void _validateUrlString(String url) {
    try {
        final Uri uri = Uri.parse(url);
        if (uri.scheme.isEmpty) {
            throw ArgumentError('Missing scheme in URL string: "$url"');
        }
    } on FormatException catch (e) {
        throw ArgumentError(e);
    }
}
